#!/usr/bin/env python3
from pathlib import Path
import click
import yaml
from click import secho

from indico_install.infra.eks import cluster, config, database, network, storage
from indico_install.config import ConfigsHolder, merge_dicts
from indico_install.utils import run_cmd, options_wrapper

SOURCES = [network, storage, database, cluster]


@click.group("eks")
@click.pass_context
def eks(ctx):
    """
    Indico infrastructure setup and validation for AWS Kubernetes Service
    """
    ctx.ensure_object(dict)
    if not ((config.access_key and config.access_secret) or config.aws_profile):
        secho("Missing AWS credentials for EKS", fg="yellow")
    else:
        secho(
            f"Using AWS profile: {config.aws_profile}"
            if config.aws_profile
            else f"Using AWS access key: {config.access_key[:5]}****",
            fg="blue",
        )


@eks.command("check")
@click.pass_context
@options_wrapper(check_input=True)
def check(ctx, *, input_yaml, **kwargs):
    """Validate all resources"""
    ctx.obj["SESSION"] = config.Session()
    if not input_yaml.is_file():
        click.secho(f"Could not find {input_yaml}.", fg="red")
        return
    failed = 0
    user_conf = ConfigsHolder(config=input_yaml)
    config.ask_for_infra_input(user_conf)
    user_conf.save()
    for resource in SOURCES:
        resource_name = resource.__name__
        secho(f"Validating {resource_name}...")
        try:
            resource.validate(user_conf, session=ctx.obj["SESSION"])
        except Exception as e:
            secho(f"{resource_name} NOT OK! Error: {e}\n", fg="red")
            failed += 1
        else:
            secho(f"{resource_name} OK!\n", fg="green")
    secho(
        f"Validation complete: {failed} errors",
        fg="red" if failed else "green",
        bold=True,
    )


@eks.command("create")
@click.pass_context
@click.argument("config_file")
@options_wrapper(check_input=True)
def create(ctx, config_file, *, input_yaml, **kwargs):
    """
    Create a new EKS cluster enabled with GPU - uses eksctl in the backend

    Args:

      <CONFIG_FILE> - the name of the config file to use for cluster configuration
    """
    if not Path(config_file).is_file():
        click.secho(f"Config file {config_file} could not be found!", fg="red")

    ctx.obj["SESSION"] = config.Session()
    run_cmd(["eksctl", "create", "cluster", "-f", config_file], tty=True)
    run_cmd(
        "kubectl create -f https://raw.githubusercontent.com/NVIDIA/k8s-device-plugin/v1.12/nvidia-device-plugin.yml"
    )
    user_conf = ConfigsHolder(config=input_yaml)
    with open(config_file, "r") as conf_file:
        conf = yaml.safe_load(conf_file)

    cluster_name = conf["metadata"]["name"]
    nodegroup = conf['nodeGroups'][0]

    if nodegroup.get("privateNetworking", False):
        user_conf["services"] = merge_dicts(
            user_conf.get("services", {}),
            {
                "app-edge": {
                    "values": {
                        "annotations": {
                            "service.beta.kubernetes.io/aws-load-balancer-internal": "0.0.0.0/0"
                        }
                    }
                }
            },
        )
    nodegroup_prefix = (
        f"eksctl-{cluster_name}-nodegroup-{nodegroup['name']}"
    )
    asg = next(
        ctx.obj["SESSION"]
        .ASGClient.get_paginator("describe_auto_scaling_groups")
        .paginate(PaginationConfig={"PageSize": 100})
        .search(f"AutoScalingGroups[?contains(AutoScalingGroupName, `{nodegroup_prefix}`)]"),
        None,
    )
    user_conf["cluster"] = user_conf.get("cluster", {}) or {}
    user_conf["cluster"].update({
        "name": conf["metadata"]["name"],
        "nodegroup_name": asg["AutoScalingGroupName"] if asg else None
    })
    user_conf.save()



from pathlib import Path
import copy
import os
import re
import tempfile
from contextlib import contextmanager

import click

from indico_install.config import (
    ConfigsLoader,
    merge_dicts,
    REMOTE_TEMPLATES_PATH,
    yaml,
)
from indico_install.setup import setup
from indico_install.utils import run_cmd, options_wrapper

var_regex = re.compile(r"<\!var:([^<>]*)>")
lookup_regex = re.compile(r"<\!lookup:([^<>]*)>")


@contextmanager
def _resolve_values(values):
    f = tempfile.NamedTemporaryFile(mode="w")
    yaml.dump(dict(values.items()), f, default_flow_style=False)
    yield f" -f {f.name}"
    f.close()


def _resolve_template(template_values, variables, conf):
    if isinstance(template_values, dict):
        new_values = {}
        for key, value in template_values.items():
            new_values[_resolve_template(key, variables, conf)] = _resolve_template(
                value, variables, conf
            )
        return new_values
    elif isinstance(template_values, list):
        return [_resolve_template(value, variables, conf) for value in template_values]
    elif isinstance(template_values, str):
        var_matches = var_regex.findall(template_values)
        lookup_matches = lookup_regex.findall(template_values)
        while var_matches or lookup_matches:
            for match in var_matches:
                parts = match.split("=", 1)
                if len(parts) > 1:
                    default = str(parts[1])
                else:
                    default = ""
                template_values = template_values.replace(
                    f"<!var:{match}>", str(variables.get(parts[0], default))
                )
            for match in lookup_matches:
                look_up_keys = match.split("|")
                target_value = dict(conf)
                parts = look_up_keys[-1].split("=", 1)
                if len(parts) > 1:
                    default = str(parts[1])
                    look_up_keys[-1] = parts[0]
                else:
                    default = ""
                for look_up_key in look_up_keys:
                    target_value = target_value.get(look_up_key, {})
                template_values = template_values.replace(
                    f"<!lookup:{match}>", str(target_value) if target_value != {} else default
                )
            var_matches = var_regex.findall(template_values)
            lookup_matches = lookup_regex.findall(template_values)

    return template_values


def helm_render(deployment_root, templates_dir, group, template, values, values_files=""):
    generated_dir = deployment_root / "generated"
    click.echo(f"Generating {values['name']}.yaml")
    with _resolve_values(values) as resolved_str:
        results = run_cmd(
            f"helm template {values_files} {templates_dir / group} --execute templates/{template}.yaml{resolved_str}",
            silent=True,
        )

    with open(generated_dir / f"{values['name']}.yaml", "w") as f:
        f.write(results)


def resolve_templates(templates):
    unresolved_templates = lambda: {
        template_name: template
        for template_name, template in templates.items()
        if "<!template>" in template
    }
    templates_to_resolve = unresolved_templates()
    while templates_to_resolve:
        for template_name, template in templates_to_resolve.items():
            template_to_merge = template.pop("<!template>")
            templates[template_name] = merge_dicts(
                copy.deepcopy(templates[template_to_merge]), templates[template_name]
            )
        templates_to_resolve = unresolved_templates()


def resolve_all(configs):
    resolve_templates(configs["_templates"])

    for resource, helm_info in configs["services"].items():
        if "<!template>" in helm_info:
            template_name = helm_info.pop("<!template>")
            configs["services"][resource] = merge_dicts(
                copy.deepcopy(configs["_templates"][template_name]), helm_info
            )
            # Attach additional values
        configs["services"][resource]["values"] = configs["services"][resource].get(
            "values", {}
        )
        configs["services"][resource] = _resolve_template(
            configs["services"][resource], configs["services"][resource], configs
        )

    configs.update(
        _resolve_template(
            {k: v for k, v in configs.items() if k not in ("_templates", "services")},
            {},
            configs,
        )
    )



def render_from_local(deployment_root, templates_dir, cluster, services_yaml, input_yaml, service=None):
    generated = deployment_root / "generated"
    click.secho(f"clearing {generated}", fg="yellow")
    os.makedirs(generated, exist_ok=True)
    run_cmd(f"rm -rf {generated/ '*.yaml'}", silent=True)

    setup(input_yaml)
    # images will be "updated" by Indico and contains cluster defaults
    # input is customer overrides
    configs = ConfigsLoader(services_yaml, input_yaml)
    resolve_all(configs)

    with _resolve_values(configs) as resolved_str:
        for resource, helm_info in configs["services"].items():
            if service and service not in resource:
                continue

            helm_info["values"] = helm_info.get("values", {})
            helm_info["values"]["name"] = resource
            configs["services"][resource] = helm_info
            if helm_info.pop("<!disabled>", None):
                continue
            extra_keys = {
                k: v
                for k, v in helm_info.items()
                if k not in ("group", "template", "values")
            }
            helm_info["values"] = merge_dicts(extra_keys, helm_info["values"])
            [helm_info.pop(k) for k in extra_keys.keys()]

            helm_render(deployment_root, templates_dir, values_files=f" {resolved_str}", **helm_info)

    configs.save(deployment_root / "configuration" / f"{cluster}.yaml")


def _resolve_remote(deployment_root, remote_path):
    """
    Download remote templates and services yaml and unpack
    Unpack if necessary, and validate contents (error if validation fails)
    Return the location of the local, unpacked templates dir
    """
    remote_templates_path = REMOTE_TEMPLATES_PATH + remote_path + "/templates.tar.gz"
    remote_services_yaml = REMOTE_TEMPLATES_PATH + remote_path + "/services.yaml"
    local_directory = deployment_root / "remote_configs" / ''.join(c for c in str(remote_path) if c.isalnum)
    click.secho(f"Downloading remote configs to {local_directory}", fg="yellow")
    if local_directory.is_dir():
        run_cmd(f"rm -rf {local_directory}", silent=True)
    os.makedirs(local_directory / "templates", exist_ok=True)

    run_cmd(f"wget {remote_templates_path} -O - | tar -xz -C {local_directory / 'templates'}")
    run_cmd(f"wget {remote_services_yaml} -O {local_directory / 'services.yaml'}")
    return (local_directory / "services.yaml", local_directory / "templates")


@click.command("render")
@click.pass_context
@click.argument("service", required=False)
@click.option("-r", "--remote-configs", help="Remote GCS folder with configs to render from. Ex: 'latest' or 'master.24235243'")
@options_wrapper(check_input=True)
def render(ctx, service=None, *, deployment_root, cluster, input_yaml, services_yaml, remote_configs, **kwargs):
    """
    Render Helm templates from the "templates" directories.

    Only render the template for services with names containing <SERVICE> if provided
    """
    deployment_root = Path(deployment_root)
    templates_dir = deployment_root / "templates"
    if remote_configs:
        services_yaml, templates_dir = _resolve_remote(deployment_root, remote_configs)

    if not services_yaml.is_file():
        click.secho(f"Could not find {services_yaml}.", fg="red")
        return

    render_from_local(deployment_root, templates_dir, cluster, services_yaml, input_yaml, service=service)
    click.secho("All set!", fg="green", bold=True)



from indico_install.utils import run_cmd


APP_MAP = {
    "review": "moonbow",
    "teach": "crowdlabel",
    "auth": "noct",
    "datasets": "sharknado",
    "discover": "elmosfire",
    "default": "default",
    "kit": "indiKit",
}


def get_app_hashes():
    output = run_cmd(
        r"""echo -e $(kubectl get configmap app-edge-nginx-conf -o json | jq '.data["nginx.conf"]') | grep -A 10 'map \$appname \$clientloc'""",
        silent=True,
    )
    output = output[output.find("{") + 1: output.find("}")].strip()
    output = output.split()

    app_hashes = {}
    for idx in range(0, len(output), 2):
        app_hash = output[idx + 1]
        app_hashes[APP_MAP[output[idx]]] = app_hash[app_hash.rfind("/") + 1: -1]
    return app_hashes


def get_non_matching_images(configs, only_first=False):
    output = run_cmd(
        "kubectl get deploy -o wide  | grep indico | awk '{print $1\"=\"$7s}'",
        silent=True,
    )
    images = {
        x[0]: ("deployment", x[1])
        for x in [y.split("=") for y in output.split("\n") if y.strip()]
    }

    output = run_cmd(
        "kubectl get statefulset -o wide  | grep indico | awk '{print $1\"=\"$5}'",
        silent=True,
    )

    images.update(
        {
            x[0]: ("statefulset", x[1])
            for x in [y.split("=") for y in output.split("\n") if y.strip()]
        }
    )
    for (deployment, (resource, image)) in images.items():
        if image.startswith("gcr.io/new-indico"):
            image = image[len("gcr.io/new-indico") + 1:]
        elif image.startswith("indicoio"):
            image = image[len("indicoio") + 1:]

        images[deployment] = (resource, image)

    for app, saved_image in configs["images"].items():
        for (deployment, (resource, cluster_image)) in images.items():
            formatted_deployment = deployment.replace("-", "").replace("_", "").lower()

            if app.lower() == formatted_deployment or app.lower() in cluster_image.replace(
                ".", ""
            ).replace(
                "-", ""
            ).replace(
                "_", ""
            ):
                if saved_image != cluster_image:
                    yield (app, resource, deployment, saved_image, cluster_image)
                if only_first:
                    break



from pathlib import Path
import click
from click import prompt

from indico_install.config import ConfigsHolder
from indico_install.utils import run_cmd, options_wrapper

SCRIPT = Path(__file__).parent / "bin" / "indico_infra"


@click.group("sn")
@click.pass_context
def single_node(ctx):
    """
    Create and validate Indico platform installation on a single node
    """
    pass


@single_node.command("check")
@click.pass_context
@options_wrapper(check_input=True)
def check(ctx, *, input_yaml, **kwargs):
    """Validate local MicroK8S cluster installation"""
    conf = ConfigsHolder(config=input_yaml)
    run_cmd([str(SCRIPT.resolve()), "check"], tty=True)

    rw_dir = Path(conf["clusterVolumes"]["rwx"]["hostPath"]["path"])
    ro_dir = Path(conf["clusterVolumes"]["rox"]["hostPath"]["path"])
    assert (
        rw_dir and rw_dir.is_dir()
    ), "Directory selected for read-write data does not exist!"
    assert (
        ro_dir and ro_dir.is_dir()
    ), "Directory selected for read-only data does not exist!"


@single_node.command("create")
@click.pass_context
@options_wrapper(check_input=True)
def create(ctx, *, input_yaml, **kwargs):
    """Install local MicroK8S cluster"""
    conf = ConfigsHolder(config=input_yaml)
    run_cmd([str(SCRIPT.resolve()), "create"], tty=True)
    ask_for_infra_input(conf)

    rw_dir = Path(conf["clusterVolumes"]["rwx"]["hostPath"]["path"])
    ro_dir = Path(conf["clusterVolumes"]["rox"]["hostPath"]["path"])

    rw_dirs = " ".join(
        [
            str(rw_dir / sub_dir)
            for sub_dir in ("postgres_data", "rainbow", "finetune_models")
        ]
    )
    run_cmd(f"sudo mkdir -pm 777 {rw_dirs} {str(ro_dir)}".split(), tty=True)


def ask_for_infra_input(conf):
    conf["clusterVolumes"]["rwx"]["hostPath"]["path"] = prompt(
        "Directory for all indico data - should have at least 800GB available",
        default=conf["clusterVolumes"]["rwx"]["hostPath"].get("path", None),
    )

    conf["clusterVolumes"]["rox"]["hostPath"]["path"] = prompt(
        "Directory for read-only-v7 API data - should have 120GB available",
        default=conf["clusterVolumes"]["rox"]["hostPath"].get("path", None),
    )
    conf.save()



from indico_install.infra.eks.config import session

from indico_install.utils import run_cmd


@session
def validate(config, session=None):
    """
    Validates the following:
    Cluster is in healthy status
    Nodes are up and connected
    ASG configuration includes:
     - sufficient volume size
     - NFS mount commands in startup scripts
    NOT IMPLEMENTED: Current machine has kubectl access
    """
    info = config.get("cluster")
    assert info, "No cluster information provided"
    cluster = session.EKSClient.describe_cluster(name=info["name"])["cluster"]
    assert cluster["status"] == "ACTIVE", "Cluster is not ACTIVE"

    try:
        asg = session.ASGClient.describe_auto_scaling_groups(
            AutoScalingGroupNames=[info["nodegroup_name"]]
        )["AutoScalingGroups"][0]
    except Exception:
        raise Exception(f"Unable to find node group of name {info['nodegroup_name']}")
    assert asg["DesiredCapacity"] >= 2, "Minimum 2 nodes required"
    assert asg["MinSize"] >= 2, "Minimum 2 nodes must be up"
    assert asg["MaxSize"] > max(
        asg["MinSize"], asg["DesiredCapacity"]
    ), "Max size of ASG must be greater than min size and desired capacity"
    assert (
        len(asg["Instances"]) == asg["DesiredCapacity"]
    ), "There are less nodes in the ASG than the desired capacity"
    assert all(
        [n["HealthStatus"] == "Healthy" for n in asg["Instances"]]
    ), "Some instances are not currently healthy"

    if asg["Instances"][0].get("LaunchConfigurationName"):
        _val_launch_config(asg, session=session)
    else:
        _val_launch_template(asg, session=session)

    # Check the cluster via kubectl
    assert (
        int(run_cmd("kubectl get nodes --no-headers | wc -l", silent=True) or "0") >= asg["DesiredCapacity"]
    ), "Some nodes have not joined the cluster"

    # Now we check to make sure GPU is enabled within the cluster
    nvidia_pods = run_cmd(
        "kubectl get pods --all-namespaces | grep 'nvidia' | grep 'Running'", silent=True
    ).strip()
    assert nvidia_pods != "", "The cluster does not have NVIDIA GPU pods running."


def _val_launch_template(asg, session):
    launch_temp_name = asg["LaunchTemplate"]["LaunchTemplateName"]

    launch_temp_ver = asg["LaunchTemplate"]["Version"]
    assert all(
        [
            n["LaunchTemplate"]["LaunchTemplateName"] == launch_temp_name
            and n["LaunchTemplate"]["Version"] == launch_temp_ver
            for n in asg["Instances"] if n["LifecycleState"] == "InService"
        ]
    ), f"Nodes are not all using the same launch template and version {launch_temp_name} , version {launch_temp_ver} - some may be out of date"

    try:
        launch_template = session.EC2Client.describe_launch_template_versions(
            LaunchTemplateName=launch_temp_name,
            Versions=[launch_temp_ver]
        )["LaunchTemplateVersions"][0]["LaunchTemplateData"]
    except Exception:
        raise AssertionError("Could not find launch template for asg")

    assert launch_template["InstanceType"].startswith("p") and launch_template[
        "InstanceType"
    ].endswith(
        "xlarge"
    ), "Instance types for ASG must be either p2 or p3, and xlarge or larger"
    assert any(
        [
            bdm["Ebs"]["VolumeSize"] >= 100
            for bdm in launch_template["BlockDeviceMappings"]
        ]
    ), "ASG launch template does not specify an EBS volume of at least 100GB for instances"


def _val_launch_config(asg, session):
    assert all(
        [
            n["LaunchConfigurationName"] == asg["LaunchConfigurationName"]
            for n in asg["Instances"]
        ]
    ), "Nodes are not all using the same launch config - some may be out of date"

    # Check how the nodes are configured
    try:
        launch_config = session.ASGClient.describe_launch_configurations(
            LaunchConfigurationNames=[asg["LaunchConfigurationName"]]
        )["LaunchConfigurations"][0]
    except Exception:
        raise AssertionError("Unable to find launch config for autoscaling group")
    assert launch_config["InstanceType"].startswith("p") and launch_config[
        "InstanceType"
    ].endswith(
        "xlarge"
    ), "Instance types for ASG must be either p2 or p3, and xlarge or larger"
    assert any(
        [
            bdm["Ebs"]["VolumeSize"] >= 100
            for bdm in launch_config["BlockDeviceMappings"]
        ]
    ), "ASG launch config does not specify an EBS volume of at least 100GB for instances"



from datetime import datetime
import click
from indico_install.utils import run_cmd, options_wrapper
from indico_install.config import ConfigsHolder, merge_dicts
from indico_install.helm import render, apply
from indico_install.setup import get_tls


@click.group("svc")
@click.pass_context
def svc(ctx):
    """Commands for cluster services"""
    pass


@svc.command("update")
@click.pass_context
@click.argument("service", required=True, type=click.Choice(["ssl"]))
@options_wrapper(check_input=True)
def update_svc(ctx, service, *, input_yaml, **kwargs):
    """
    Function to update an existing service.

    Args:

      <SERVICE>: The service to update. Choices are:

        - ssl : update certificates
    """
    if service == "ssl":
        config = ConfigsHolder(config=input_yaml)
        get_tls(config, redo=True)
        ctx.invoke(render, input_yaml=input_yaml, **kwargs)
        ctx.invoke(
            apply, service="app-edge-nginx-conf", input_yaml=input_yaml, **kwargs
        )
        ctx.invoke(restart, service="app-edge")
    else:
        click.secho(f"Service {service} has no update logic.")


@svc.command("scale")
@click.pass_context
@click.argument("service", required=True, type=str)
@click.argument("amount", required=True, type=int)
@options_wrapper()
def scale(ctx, service, amount, *, input_yaml, yes, **kwargs):
    """
    Scale a K8S cluster deployment or statefulset

    ARGS:
        <SERVICE> grep string of deployments and statefulsets to scale

        <AMOUNT> number of pods to create
    """
    updated_svcs = []
    for svc_type in ["deployment", "statefulset"]:
        out = run_cmd(
            """kubectl get %s --no-headers | grep "%s" | awk '{print $1}'"""
            % (svc_type, service),
            silent=True,
        )
        if not out:
            continue
        for _svc in out.splitlines():
            click.secho(
                run_cmd(
                    """kubectl patch %s %s -p '{"spec":{"replicas":%d}}'"""
                    % (svc_type, _svc, amount)
                ),
                fg="green",
            )
            updated_svcs.append(_svc)

    if updated_svcs and yes:
        if not input_yaml.is_file():
            click.secho(
                f"Could not find {input_yaml}. Unable to save new scale", fg="yellow"
            )
            return
        conf = ConfigsHolder(config=input_yaml)
        updated_svcs_dict = {
            _svc: {"values": {"replicas": amount}} for _svc in updated_svcs
        }
        conf["services"] = merge_dicts(conf.get("services", {}), updated_svcs_dict)
        conf.save()


@svc.command("restart")
@click.pass_context
@click.argument("service", required=True)
def restart(ctx, service):
    """
    Reroll a K8S cluster deployment or statefulset.

    ARGS:

        <SERVICE> grep string of deployments and statefulsets to reroll

         - "all" reset all deployments and statefulsets with an inditype label of service or celerytask
    """
    patch_string = '{"spec":{"template":{"metadata":{"labels":{"date":"%d"}}}}}' % (
        int(datetime.now().timestamp())
    )
    for svc_type in ["deployment", "statefulset"]:
        if service == "all":
            out = run_cmd(
                """kubectl get %s -l 'inditype in (service, celerytask)' --no-headers | awk '{print $1}' | xargs -n 1 -I {} kubectl patch %s {} -p '%s'"""
                % (svc_type, svc_type, patch_string),
                silent=True,
            )
        else:
            out = run_cmd(
                """kubectl get %s --no-headers | grep "%s" | awk '{print $1}' | xargs -n 1 -I {} kubectl patch %s {} -p '%s'"""
                % (svc_type, service, svc_type, patch_string),
                silent=True,
            )
        if out:
            click.secho(out, fg="green")
